﻿#include "Interactive.h"


/*********************************************************************************************
* 功能    ：	选择功能码
* 描述	  ：	在功能码 0x01 0x03 0x0F 0x10中选择一个功能码
* 输入    ：    无
* 输出    ：    Function_code  选择的功能码
*********************************************************************************************/
int Input_Function_code(void)
{
	int function_code;
	cin >> hex;
	int i = 1;
	printf("功能码 ：  0x01      0x03        0x0F         0x10 \n");
	printf("说明   ： 读线圈   读寄存器   写多个线圈  写多个寄存器\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入选择的功能码：";
		}
		else
			cout << "输入功能码错误，请重新选择的功能码";
		cin >> function_code;
		cin.clear();
		cin.sync();
	} while (function_code != 0x01 && function_code != 0x03 && function_code != 0x0F && function_code != 0x10);
	return function_code;
}

/*********************************************************************************************
* 功能    ：	输入操作起始地址码
* 描述	  ：	起始地址码，最大值65535
* 输入    ：    无
* 输出    ：    starting_address 操作起始地址码
*********************************************************************************************/
unsigned int Input_Starting_address(void)
{
	unsigned int starting_address;
	int i = 1;
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)起始地址：";
		}
		else
			cout << "请重新输入操作(十进制)起始地址";
		cin >> starting_address;
		cin.clear();
		cin.sync();
	} while (starting_address > 65535 || starting_address < 0);
	return starting_address;
}

/*********************************************************************************************
* 功能    ：	输入操作数量
* 描述	  ：	操作数量，0x01 1-200  0x03 1-125  0x0xF 1-1968 0x10 1-123
* 输入    ：    function_code 选择的功能码类型
* 输出    ：    operations_Number 对应功能码类型的操作数量
*********************************************************************************************/
unsigned int Input_Operations_number(int function_code)
{
	unsigned int operations_Number;
	int i = 1;
	bool flage = true;
	printf("  功能码  0x01    0x03    0x0F   0x10\n");
	printf("操作数量  1-2000  1-125  1-1968  1-123\n");
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)数量：";
		}
		else
			cout << "请重新输入操作(十进制)数量";
		cin >> operations_Number;
		cin.clear();
		cin.sync();
		if (function_code == 0x01 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x03 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x0F && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x10 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
	} while (flage);
	return operations_Number;
}

/*********************************************************************************************
* 功能    ：	检测写入数据
* 描述	  ：	检测写入数据格式是否正确
* 输入    ：    write_date 写入的数据 write_date_number写入数据的组数
* 输出    ：    true  写入数据格式正确
*				false 写入数据格式错误
*********************************************************************************************/
bool Check_Write_date(string write_date, int write_date_number)
{
	int write_date_len = 2 * write_date_number + write_date_number - 1;
	if (write_date.length() != write_date_len)
		return false;
	int space_number = 0;
	int temp = -1;
	for (unsigned int i = 0; i < write_date.length(); i++)
	{
		if (write_date[i] == 0x20)//空格位置判定
		{
			if (i - temp == 3)
			{
				space_number++;
				temp = i;
				continue;
			}
			else
				return false;
		}
		if ((write_date[i] >= 0x30 && write_date[i] <= 0x39)) // 1-9
			continue;
		else if ((write_date[i] >= 0x41 && write_date[i] <= 0x5A)) //A-Z
			continue;
		else if ((write_date[i] >= 0x61 && write_date[i] <= 0x7A)) //a-z
			continue;
		else
			return false;
	}
	if (space_number != write_date_number - 1)
		return false;
	return true;
}

/*********************************************************************************************
* 功能    ：	计算写入数据的字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    function_code 选择的功能码类型 operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date_number 写入数据的字节数
*********************************************************************************************/
unsigned int Count_Write_date_number(int function_code, unsigned int operations_number)
{
	unsigned int write_date_number = 0;
	if (function_code == 0x01 || function_code == 0x03)
		return write_date_number;
	if (function_code == 0x0F) //读线圈
	{
		write_date_number = operations_number / 8;
		if (operations_number % 8)
			write_date_number++;
	}
	if (function_code == 0x10)
		write_date_number = operations_number * 2;
	return write_date_number;
}


/*********************************************************************************************
* 功能    ：	输入写入数据
* 描述	  ：	根据写入数量和功能码类型输入相应的数据
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date 写入的数据
*********************************************************************************************/
string Input_Write_date(int function_code, unsigned int operations_Number)
{
	int i = 1;
	string write_date;
	unsigned int write_date_number = Count_Write_date_number(function_code, operations_Number);
	if (function_code == 0x01 || function_code == 0x03)
		return write_date;//读线圈/寄存器不需要输入写入数据

	printf("请输入N组写入数据（2位16进制为一组，空格间隔）例如：00 00 01 00 20 05\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			printf("请输入%d组写入数据 ：", write_date_number);
		}
		else
			printf("请重新输入%d组写入数据 ：", write_date_number);
		getline(cin, write_date);
		cin.clear();
		cin.sync();
	} while (!Check_Write_date(write_date, write_date_number));

	return write_date;
}


/*********************************************************************************************
* 功能    ：	打印读取从站的线圈状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Coil_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	unsigned int number = ((Request_Message[10] << 8) | Request_Message[11]);
	printf("线圈起始地址为%d \n", temp1);
	unsigned int temp = temp1;
	for (int i = 0; i < Response_Message[8]; i++)
	{
		unsigned int temp2 = temp + 7;
		if (temp2 > temp1 + number - 1)
			temp2 = temp1 + number - 1;
		printf("线圈%d --- %d的状态为：%02X \n", temp2, temp, Response_Message[9 + i]);
		temp = temp + 8;
	}
}

/*********************************************************************************************
* 功能    ：	打印读取从站的寄存器状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Register_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	printf("寄存器起始地址为%d \n", temp1);
	for (int i = 0; i < Response_Message[8]; i = i + 2)
	{
		printf("寄存器%d的值为：%02X %02X \n", temp1++, Response_Message[9 + i], Response_Message[10 + i]);
	}
}

/*********************************************************************************************
* 功能    ：	打印读取异常响应报文信息
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Anomaly_date(UINT8 *Response_Message)
{
	UINT8 anomaly_code = Response_Message[8];
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	switch (anomaly_code)
	{
	case 0x01:	printf("%02X : 从站设备不支持此功能码\n", anomaly_code); break;
	case 0x02:	printf("%02X : 指定的数据地址在从站设备中不存在\n", anomaly_code); break;
	case 0x03:	printf("%02X : 指定的数据超过范围或者不允许使用\n", anomaly_code); break;
	case 0x04:	printf("%02X : 从站设备处理响应的过程中，出现未知错误等\n", anomaly_code); break;
	default:	printf("Unkown Other Error!!!!!\n");
	}
}

/*********************************************************************************************
* 功能    ：	日志记录
* 描述	  ：	记录每次通信的请求和响应报文
* 输入    ：    *Message 消息帧存放的数组   flage = 1 请求报文  flage = 其他 响应报文
*				Message_len 报文数据长度
* 输出    ：    无
*********************************************************************************************/
void Log_Note(UINT8 *Message, int flage, int message_len)
{
	FILE *fp = NULL;
	time_t timep;
	struct tm *p;
	time(&timep);
	p = gmtime(&timep);
	string recv_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Recv:";
	string send_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Send:";

	/* 打开文件用于读写 */
	if ((fp = fopen("Modbus_TCP_log.txt", "a+")) == NULL)
	{
		printf("打开文件失败");
		return;
	}
	if (flage == 1)
	{
		fwrite(send_str.c_str(), send_str.length(), 1, fp);
		for (int i = 0; i < message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	else
	{
		fwrite(recv_str.c_str(), recv_str.length(), 1, fp);
		for (int i = 0; i < message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
	fp = NULL;
}

void Printf_Message(UINT8 *Message, int flage, int message_len)
{
	if (flage == 1)
		printf("主站请求 ：");
	else
		printf("从站响应 ：");
	for (int i = 0; i < message_len; i++)
	{
		printf("%02x  ", Message[i]);
	}
	printf("\n");
	if (LOG_NOTE_SWITCH)
		Log_Note(Message, flage, message_len);
}

void Printf_Analysis_Outcome(UINT8 *Response_Message, UINT8 *Request_Message)
{
	switch (Response_Message[7])
	{
	case 0x01:Printf_Coil_date(Response_Message, Request_Message); break;
	case 0x03:Printf_Register_date(Response_Message, Request_Message); break;
	case 0x0F:printf("成功写入从站线圈%d个\n", ((Request_Message[10] << 8) | Request_Message[11])); break;
	case 0x10:printf("成功写入从站寄存器%d个\n", ((Request_Message[10] << 8) | Request_Message[11])); break;
	default: printf("Other Error");
	}
}

void Printf_Interactive_Data(UINT8 *Response_Message, UINT8 *Request_Message, int flage, int request_message_len, int response_message_len)
{
	Printf_Message(Request_Message, 1, request_message_len);
	Printf_Message(Response_Message, 0, response_message_len);
	switch (flage)
	{
	case MSG_LEN_ERROR: printf("Message Length Error"); break;
	case MABP_ERROR:printf("MBAP Date Error"); break;
	case FUNCTION_CODE_ERROR:printf("Function Code Error"); break;
	case START_ADDRESS_ERROR:printf("Start Address Error"); break;
	case OPERATION_NUMBER_ERROR:printf("Operation Number Error"); break;
	case NORMAL_RESPONSE: Printf_Analysis_Outcome(Response_Message, Request_Message); break;
	case ABNORMAL_RESPONSE:Printf_Anomaly_date(Response_Message); break;
	default:printf("Other Error");
	}
}